from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Iterable, Literal, overload


if TYPE_CHECKING:
    from mipa.ext.timelines.core import AbstractTimeline
    from aiohttp.client_ws import ClientWebSocketResponse

__all__ = ["Router"]

IChannel = Literal["global", "main", "home", "local", "hybrid"]


CHANNELS = {
    "global": "globalTimeline",
    "main": "main",
    "home": "homeTimeline",
    "local": "localTimeline",
    "hybrid": "hybridTimeline",
}


class Router:
    def __init__(
        self, web_socket: ClientWebSocketResponse, max_capure: int = 100
    ):
        self.web_socket: ClientWebSocketResponse = web_socket
        self.captured_note: list[str] = []
        self.max_capture: int = max_capure
        self.__channel_ids: dict[str, IChannel] = {}
        self.__channel_handlers: dict[str, AbstractTimeline] = {}

    @overload
    async def connect_channel(self, channel_list: Iterable[IChannel]):
        ...

    @overload
    async def connect_channel(
        self, channel_list: dict[IChannel, AbstractTimeline | None]
    ):
        ...

    async def connect_channel(
        self,
        channel_list: Iterable[IChannel]
        | dict[IChannel, AbstractTimeline | None],
    ):
        """
        Connects to a channel based on the list passed.

        Parameters
        ----------
        channel_list : IChannel
            ['global', 'main', 'home', 'local', 'hybrid']

        Returns
        -------
        dict[IChannel, str]
        """

        _channel_ids: dict[IChannel, str] = {}
        try:
            for channel in channel_list:
                channel_id = f"{uuid.uuid4()}"
                _channel_ids[channel] = channel_id
                self.__channel_ids[channel_id] = channel
                if isinstance(channel_list, dict):
                    channel_handler = channel_list[channel]
                    if channel_handler:
                        self.__channel_handlers[channel_id] = channel_handler

                await self.web_socket.send_json(
                    {
                        "type": "connect",
                        "body": {
                            "channel": f"{CHANNELS[channel]}",
                            "id": f"{_channel_ids[channel]}",
                        },
                    }
                )
        except KeyError:
            pass
        return _channel_ids

    async def disconnect_channel(self, channel_id: str):
        """
        Disconnects from a channel based on the id passed.

        Parameters
        ----------
        channel_id : str
        """

        self.__channel_ids.pop(channel_id)
        await self.web_socket.send_json(
            {"type": "disconnect", "body": {"id": f"{channel_id}"}}
        )

    async def capture_message(self, note_id: str) -> None:
        """
        Captures a message based on the id passed.
        Parameters
        ----------
        note_id : str
        """
        if len(self.captured_note) > self.max_capture:
            await self.web_socket.send_json(
                {"type": "unsubNote", "body": {"id": f"{note_id}"}}
            )
            del self.captured_note[0]
        self.captured_note.append(note_id)
        await self.web_socket.send_json(
            {"type": "subNote", "body": {"id": f"{note_id}"}}
        )

    @property
    def channel_ids(self) -> dict[str, IChannel]:
        """
        Returns the unique ID of the connected channel and the channel mapping

        Returns
        -------
        dict[str, IChannel]
        """
        return self.__channel_ids

    @property
    def channel_handlers(self) -> dict[str, AbstractTimeline]:
        """
        Returns the unique ID of the connected channel and the channel mapping

        Returns
        -------
        dict[str, AbstractTimeline]
        """
        return self.__channel_handlers



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.
    - Simplification of some functions.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, TypeVar

import aiohttp
from aiohttp import ClientError
from mipac.utils.format import str_lower

from mipa.exception import ClientConnectorError, WebSocketReconnect
from mipa.router import Router

if TYPE_CHECKING:
    from .client import Client

__all__ = ("MisskeyWebSocket",)


MS = TypeVar("MS", bound="aiohttp.ClientWebSocketResponse")


class MisskeyWebSocket:
    def __init__(self, socket: MS, client: Client):
        self.socket: MS = socket
        self._dispatch = lambda *args: None
        self._connection = None
        self.client = client
        self._misskey_parsers: Optional[Dict[str, Callable[..., Any]]] = None

    @classmethod
    async def from_client(
        cls, client: Client, *, timeout: int = 60, event_name: str = "ready"
    ):
        try:
            socket = await client.core.http.session.ws_connect(
                f"{client.url}?i={client.token}"
            )
            ws = cls(socket, client)
            ws._dispatch = client.dispatch
            ws._connection = client._connection
            ws._misskey_parsers = client._connection.parsers
            client._router = Router(socket, max_capure=client.max_capture)
            client.dispatch(event_name, socket)
            return ws
        except (ClientConnectorError, ClientError):
            while True:
                await asyncio.sleep(3)
                return await cls.from_client(
                    client, timeout=timeout, event_name=event_name
                )

        # await ws.poll_event(timeout=timeout)

    async def received_message(self, msg, /):
        if isinstance(msg, bytes):
            msg = msg.decode()

        await self._misskey_parsers[str_lower(msg["type"]).upper()](msg)

    async def poll_event(self, *, timeout: int = 60):
        msg = await self.socket.receive(timeout=timeout)

        if msg is aiohttp.http.WS_CLOSED_MESSAGE:
            raise WebSocketReconnect()
        elif msg is aiohttp.http.WS_CLOSING_MESSAGE:
            raise WebSocketReconnect()
        elif msg.type is aiohttp.WSMsgType.TEXT:
            await self.received_message(json.loads(msg.data))
        elif msg.type is aiohttp.WSMsgType.ERROR:
            raise WebSocketReconnect()



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.
    - Adding new functions and methods

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import importlib
import inspect
import re
import sys
import traceback
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    List,
    Optional,
    Tuple,
    Union,
)

from mipac.models.emoji import CustomEmoji
from mipac.models.user import MeDetailed, UserDetailedNotMe

from mipa import Client
from mipa.exception import (
    CogNameDuplicate,
    ExtensionAlreadyLoaded,
    ExtensionFailed,
    InvalidCogPath,
    NoEntryPointError,
)
from mipa.ext.commands.context import Context
from mipa.ext.commands.core import CommandManager

if TYPE_CHECKING:
    from aiohttp.client_ws import ClientWebSocketResponse
    from mipac.models.notification import (
        NotificationAchievement,
        NotificationFollow,
        NotificationFollowRequest,
        NotificationNote,
        NotificationPollEnd,
        NotificationReaction,
    )

    from mipa.ext import Cog


__all__ = ["BotBase", "Bot"]


class BotBase(CommandManager):
    def __init__(self, **options: dict[Any, Any]):
        super().__init__(**options)
        self.extra_events: dict[str, Any] = {}
        self.special_events: dict[str, Any] = {}
        self._check_once: List[Any] = []  # TODO: いつか確認する
        self._checks: List[Any] = []  # TODO: いつか確認する
        self._after_invoke = None
        self.token: Optional[str] = None
        self.origin_uri: Optional[str] = None
        self.__extensions: dict[str, Any] = {}
        self.user: MeDetailed
        self.__cogs: dict[str, Cog] = {}
        self.strip_after_prefix = options.get("strip_after_prefix", False)
        # self.logger = get_module_logger(__name__) TODO: 直す
        self.loop = asyncio.get_event_loop()

    def _on_note(self, message):
        self.dispatch("note", message)

    async def on_ready(self, ws: ClientWebSocketResponse):
        """
        on_readyのデフォルト処理

        Parameters
        ----------
        ws : ClientWebSocketResponse
        """

    def event(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_event(func, name)
            return func

        return decorator

    def add_event(
        self, func: Coroutine[Any, Any, Any], name: Optional[str] = None
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Listeners must be coroutines")

        if name in self.extra_events:
            self.special_events[name].append(func)
        else:
            self.special_events[name] = [func]

    def listen(self, name: Optional[str] = None):
        def decorator(func: Coroutine[Any, Any, Any]):
            self.add_listener(func, name)
            return func

        return decorator

    def add_listener(
        self,
        func: Union[Coroutine[Any, Any, Any], Callable[..., Any]],
        name: Optional[str] = None,
    ):
        name = func.__name__ if name is None else name
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Listeners must be coroutines")

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    async def event_dispatch(
        self, event_name: str, *args: Tuple[Any], **kwargs: dict[Any, Any]
    ) -> bool:
        """
        on_ready等といった

        Parameters
        ----------
        event_name :
        args :
        kwargs :

        Returns
        -------

        """
        ev = f"on_{event_name}"
        for event in self.special_events.get(ev, []):
            foo = importlib.import_module(event.__module__)
            coro = getattr(foo, ev)
            await self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            await self.schedule_event(getattr(self, ev), ev, *args, **kwargs)
        return ev in dir(self)

    def dispatch(
        self, event_name: str, *args: tuple[Any], **kwargs: dict[Any, Any]
    ):
        ev = f"on_{event_name}"
        for event in self.extra_events.get(ev, []):
            if inspect.ismethod(event):
                coro = event
                event = event.__name__
            else:
                foo = importlib.import_module(event.__module__)
                coro = getattr(foo, ev)
            self.schedule_event(coro, event, *args, **kwargs)
        if ev in dir(self):
            self.schedule_event(getattr(self, ev), ev, *args, **kwargs)

    async def add_cog(self, cog: Cog, override: bool = False) -> None:
        cog_name = cog.__cog_name__
        existing = self.__cogs.get(cog_name)
        if existing is not None:
            if not override:
                raise CogNameDuplicate()
            await self.remove_cog(cog_name)  # TODO: 作る

        cog = cog._inject(self)
        self.__cogs[cog_name] = cog

    async def remove_cog(self, name: str):  # TODO: Optional[Cog]を返すように
        """Cogを削除します"""
        cog = self.__cogs.get(name)
        if cog is None:
            return

        cog._inject(self)

        return cog

    async def _load_from_module(self, spec: ModuleType, key: str) -> None:
        try:
            setup = spec.setup
        except AttributeError as e:
            raise NoEntryPointError(f"{key} にsetupが存在しません") from e

        try:
            await setup(self)
        except Exception as e:
            raise ExtensionFailed(key, e) from e
        else:
            self.__extensions[key] = spec

    @staticmethod
    def _resolve_name(name: str, package: Optional[str]) -> str:
        try:
            return importlib.util.resolve_name(name, package)  # pyright: ignore
        except ImportError as e:
            raise InvalidCogPath(name) from e

    async def load_extension(
        self, name: str, *, package: Optional[str] = None
    ) -> None:
        """拡張をロードする

        Parameters
        ----------
        name : str
            [description]
        package : Optional[str], optional
            [description], by default None
        """
        name = self._resolve_name(name, package)
        if name in self.__extensions:
            raise ExtensionAlreadyLoaded
        try:
            module = importlib.import_module(name)
        except ModuleNotFoundError as e:
            raise InvalidCogPath(f"cog: {name} へのパスが無効です") from e
        await self._load_from_module(module, name)

    def schedule_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: tuple[Any],
        **kwargs: dict[Any, Any],
    ) -> asyncio.Task[Any]:
        return asyncio.create_task(
            self._run_event(coro, event_name, *args, **kwargs),
            name=f"MiPA: {event_name}",
        )

    async def _run_event(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.__on_error(event_name)
            except asyncio.CancelledError:
                pass

    @staticmethod
    async def __on_error(event_method: str) -> None:
        print(f"Ignoring exception in {event_method}", file=sys.stderr)
        traceback.print_exc()

    async def on_error(self, err):
        await self.event_dispatch("error", err)

    def get_cog(self, name: str) -> Cog | None:
        return self.__cogs.get(name)

    async def get_context(self, message, cmd, cls=Context) -> Context:
        return cls(message=message, bot=self, cmd=cmd)

    async def progress_command(self, message):
        for cmd in self.all_commands:
            ctx = await self.get_context(message, cmd)
            if cmd.cmd_type == "regex":
                if re.search(cmd.key, message.content):
                    hit_list = re.findall(cmd.key, message.content)
                    if isinstance(hit_list, list):
                        hit_list = tuple(hit_list)

                    if isinstance(hit_list[0], tuple):
                        hit_list = tuple(
                            i for i in hit_list[0] if len(i.rstrip()) > 0
                        )
                    ctx.args = hit_list
                    await cmd.func.invoke(ctx)
            elif message.content.find(cmd.key) != -1:
                await cmd.func.invoke(ctx)
            else:
                continue

    async def on_user_follow(self, user: UserDetailedNotMe):
        """
        When you follow a user

        Parameters
        ----------
        user : UserDetailed
        """

    async def on_user_unfollow(self, user: UserDetailedNotMe):
        """
        When you unfollow a user

        Parameters
        ----------
        user : UserDetailed
        """

    async def on_user_followed(self, notice: NotificationFollow):
        """
        When someone follows you

        Parameters
        ----------
        notice : NotificationFollow
        """

    async def on_mention(self, notice: NotificationNote):
        """
        When someone mentions you

        Parameters
        ----------
        notice : NotificationNote
        """
        await self.progress_command(notice.note)

    async def on_reply(self, notice: NotificationNote):
        """
        When someone replies to you

        Parameters
        ----------
        notice : NotificationNote
        """

    async def on_renote(self, notice: NotificationNote):
        """
        When someone renote your note

        Parameters
        ----------
        notice : NotificationNote
        """

    async def on_quote(self, notice: NotificationNote):
        """
        When someone quote your note

        Parameters
        ----------
        notice : NotificationNote
        """

    async def on_reaction(self, notice: NotificationReaction):
        """
        When someone react to your note

        Parameters
        ----------
        notice : NotificationReaction
        """

    async def on_poll_vote(self, notice: NotificationNote):
        """
        When someone vote to your poll

        Parameters
        ----------
        notice : NotificationNote
        """

    async def on_poll_end(self, notice: NotificationPollEnd):
        """
        When a poll is ended

        Parameters
        ----------
        notice : NotificationPollEnd
        """

    async def on_follow_request(self, notice: NotificationFollowRequest):
        """
        When someone send you a follow request

        Parameters
        ----------
        notice : NotificationFollowRequest
        """

    async def on_follow_request_accept(self, notice: NotificationFollow):
        """
        When someone accept your follow request

        Parameters
        ----------
        notice : NotificationFollow
        """


    async def on_achievement_earned(self, notice: NotificationAchievement):
        """
        When you earn an achievement

        Parameters
        ----------
        notice : NotificationAchievement
        """

    async def on_emoji_deleted(self, emojis: list[CustomEmoji]):
        """
        カスタム絵文字が削除された

        Parameters
        ----------
        emojis : list[CustomEmoji]
            削除された絵文字のリスト
        """

    async def on_emoji_updated(self, emojis: list[CustomEmoji]):
        """
        カスタム絵文字が更新された

        Parameters
        ----------
        emojis : list[CustomEmoji]
            更新された絵文字のリスト
        """


class Bot(BotBase, Client):
    pass



from ._version import get_versions

__title__ = "mipa"
__author__ = "yupix"
__license__ = "MIT"
__copyright__ = "Copyright 2022-present yupix"
__author_email__ = "yupi0982@outlook.jp"
__version__ = get_versions()["version"]

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

del get_versions

from .client import Client
from .ext import *

__all__ = ("Client",)

from . import _version

__version__ = _version.get_versions()["version"]



__all__ = (
    "MIPABaseException",
    "MIPABaseWebSocketError",
    "WebSocketNotConnected",
    "WebSocketReconnect",
    "CogNameDuplicate",
    "ExtensionAlreadyLoaded",
    "ExtensionFailed",
    "InvalidCogPath",
    "NoEntryPointError",
    "ClientConnectorError",
    "TaskNotRunningError",
)


class MIPABaseException(Exception):
    """MIPA Base Exception"""


class MIPABaseWebSocketError(MIPABaseException):
    """Websocket Base Exception"""


class WebSocketNotConnected(MIPABaseWebSocketError):
    """Websocket not connected"""


class WebSocketReconnect(MIPABaseWebSocketError):
    """Websocket should reconnect"""


class CogNameDuplicate(MIPABaseException):
    """Cogの名前が重複している"""


class ExtensionNotLoaded(MIPABaseException):
    """Cogが読み込まれていない"""


class ExtensionAlreadyLoaded(MIPABaseException):
    """Cogは既に読み込まれている"""


class ExtensionFailed(MIPABaseException):
    """Cog周りのエラー"""


class InvalidCogPath(MIPABaseException):
    """無効なCogのパス"""


class NoEntryPointError(MIPABaseException):
    """Cogにsetup関数が存在しない"""


class ClientConnectorError(MIPABaseException):
    """WebSocketの接続に問題が発生した"""


class TaskNotRunningError(MIPABaseException):
    """タスクが動いてない状態で停止しようとした"""



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

The Software is modified as follows:
    - Delete unused functions and method.
    - Removing functions beyond what is necessary to make it work.
    - Simplification of some functions.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import asyncio
import inspect
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
)

from mipa.exception import TaskNotRunningError
from mipa.utils import MISSING

__all__ = ["Loop", "loop"]

_func = Callable[..., Coroutine[Any, Any, Any]]
LF = TypeVar("LF", bound=_func)
T = TypeVar("T")


class Loop(Generic[LF]):
    def __init__(
        self,
        coro: LF,
        seconds: float,
        count: Optional[int],
    ):
        self.seconds = seconds
        self.coro: LF = coro
        self.count: Optional[int] = count
        self._current_loop = 0
        self._task: Optional[asyncio.Task[None]] = None
        self._injected = None

        self._stop_next_iteration = False

        if self.count is not None and self.count <= 0:
            raise ValueError("count must be greater than 0 or None.")

        if not inspect.iscoroutinefunction(self.coro):
            raise TypeError(
                f"Expected coroutine function, not {type(self.coro).__name__!r}."
            )

    def start(
        self, *args: tuple[Any], **kwargs: Dict[Any, Any]
    ) -> asyncio.Task[Any]:
        """
        タスクを開始する

        Parameters
        ----------
        args : Any
        kwargs : Any

        Returns
        -------
        _task : asyncio.Task[Any]
        """
        if self._injected is not None:
            args = (self._injected, *args)
        self._task = asyncio.create_task(self._loop(*args, **kwargs))
        return self._task

    def stop(self):
        """
        タスクを停止
        """

        if self._task is None:
            raise TaskNotRunningError("タスクは起動していません")

        if not self._task.done():
            self._stop_next_iteration = True

    async def _loop(self, *args: tuple[Any], **kwargs: Dict[Any, Any]):
        while True:
            if self._stop_next_iteration is True:
                return
            await self.coro(*args, **kwargs)
            await asyncio.sleep(self.seconds)

            self._current_loop += 1
            if self._current_loop == self.count:
                break

    def __get__(self, obj: T, objtype: Type[T]):
        if obj is None:
            return self

        copy: Loop[LF] = Loop(
            self.coro,
            seconds=self.seconds,
            count=self.count,
        )
        copy._injected = obj

        setattr(obj, self.coro.__name__, copy)
        return copy


def loop(
    *,
    seconds: float = MISSING,
    count: Optional[int] = None,
) -> Callable[[LF], Loop[LF]]:
    def decorator(func: LF) -> Loop[LF]:
        return Loop[LF](
            func,
            seconds=seconds,
            count=count,
        )

    return decorator
